@@ -1,12 +1,13 @@
 import collections
 import os
 import subprocess
+import sys
 from pathlib import Path
 from typing import Dict
 from unittest.mock import Mock
 
 import pytest
-from pylsp import uris
+from pylsp import _utils, uris
 from pylsp.config.config import Config
 from pylsp.workspace import Document, Workspace
 
@@ -16,11 +17,12 @@ from pylsp_mypy import plugin
 DOC_URI = f"file:/{Path(__file__)}"
 DOC_TYPE_ERR = """{}.append(3)
 """
-TYPE_ERR_MSG = '"Dict[<nothing>, <nothing>]" has no attribute "append"  [attr-defined]'
+TYPE_ERR_MSG = '"Dict[<nothing>, <nothing>]" has no attribute "append"'
 
-TEST_LINE = 'test_plugin.py:279:8: error: "Request" has no attribute "id"'
-TEST_LINE_WITHOUT_COL = "test_plugin.py:279: " 'error: "Request" has no attribute "id"'
-TEST_LINE_WITHOUT_LINE = "test_plugin.py: " 'error: "Request" has no attribute "id"'
+TEST_LINE = 'test_plugin.py:279:8:279:16: error: "Request" has no attribute "id"  [attr-defined]'
+TEST_LINE_NOTE = (
+    'test_plugin.py:124:1:129:77: note: Use "-> None" if function does not return a value'
+)
 
 windows_flag: Dict[str, int] = (
     {"creationflags": subprocess.CREATE_NO_WINDOW} if os.name == "nt" else {}  # type: ignore
@@ -43,63 +45,53 @@ def workspace(tmpdir):
 
 
 class FakeConfig(object):
-    def __init__(self):
-        self._root_path = "C://" if os.name == "nt" else "/"
+    def __init__(self, path):
+        self._root_path = path
 
     def plugin_settings(self, plugin, document_path=None):
         return {}
 
 
-def test_settings():
-    config = FakeConfig()
+def test_settings(tmpdir):
+    config = Config(uris.from_fs_path(str(tmpdir)), {}, 0, {})
     settings = plugin.pylsp_settings(config)
     assert settings == {"plugins": {"pylsp_mypy": {}}}
 
 
 def test_plugin(workspace, last_diagnostics_monkeypatch):
-    config = FakeConfig()
     doc = Document(DOC_URI, workspace, DOC_TYPE_ERR)
-    plugin.pylsp_settings(config)
-    diags = plugin.pylsp_lint(config, workspace, doc, is_saved=False)
+    plugin.pylsp_settings(workspace._config)
+    diags = plugin.pylsp_lint(workspace._config, workspace, doc, is_saved=False)
 
     assert len(diags) == 1
     diag = diags[0]
     assert diag["message"] == TYPE_ERR_MSG
     assert diag["range"]["start"] == {"line": 0, "character": 0}
-    assert diag["range"]["end"] == {"line": 0, "character": 1}
+    # Running mypy in 3.7 produces wrong error ends this can be removed when 3.7 reaches EOL
+    if sys.version_info < (3, 8):
+        assert diag["range"]["end"] == {"line": 0, "character": 1}
+    else:
+        assert diag["range"]["end"] == {"line": 0, "character": 9}
+    assert diag["severity"] == 1
+    assert diag["code"] == "attr-defined"
 
 
 def test_parse_full_line(workspace):
     diag = plugin.parse_line(TEST_LINE)  # TODO parse a document here
     assert diag["message"] == '"Request" has no attribute "id"'
     assert diag["range"]["start"] == {"line": 278, "character": 7}
-    assert diag["range"]["end"] == {"line": 278, "character": 8}
+    assert diag["range"]["end"] == {"line": 278, "character": 16}
+    assert diag["severity"] == 1
+    assert diag["code"] == "attr-defined"
 
 
-def test_parse_line_without_col(workspace):
-    doc = Document(DOC_URI, workspace)
-    diag = plugin.parse_line(TEST_LINE_WITHOUT_COL, doc)
-    assert diag["message"] == '"Request" has no attribute "id"'
-    assert diag["range"]["start"] == {"line": 278, "character": 0}
-    assert diag["range"]["end"] == {"line": 278, "character": 1}
-
-
-def test_parse_line_without_line(workspace):
-    doc = Document(DOC_URI, workspace)
-    diag = plugin.parse_line(TEST_LINE_WITHOUT_LINE, doc)
-    assert diag["message"] == '"Request" has no attribute "id"'
-    assert diag["range"]["start"] == {"line": 0, "character": 0}
-    assert diag["range"]["end"] == {"line": 0, "character": 6}
-
-
-@pytest.mark.parametrize("word,bounds", [("", (7, 8)), ("my_var", (7, 13))])
-def test_parse_line_with_context(monkeypatch, word, bounds, workspace):
-    doc = Document(DOC_URI, workspace)
-    monkeypatch.setattr(Document, "word_at_position", lambda *args: word)
-    diag = plugin.parse_line(TEST_LINE, doc)
-    assert diag["message"] == '"Request" has no attribute "id"'
-    assert diag["range"]["start"] == {"line": 278, "character": bounds[0]}
-    assert diag["range"]["end"] == {"line": 278, "character": bounds[1]}
+def test_parse_note_line(workspace):
+    diag = plugin.parse_line(TEST_LINE_NOTE)
+    assert diag["message"] == 'Use "-> None" if function does not return a value'
+    assert diag["range"]["start"] == {"line": 123, "character": 0}
+    assert diag["range"]["end"] == {"line": 128, "character": 77}
+    assert diag["severity"] == 3
+    assert diag["code"] is None
 
 
 def test_multiple_workspaces(tmpdir, last_diagnostics_monkeypatch):
@@ -108,20 +100,21 @@ def foo():
     return
     unreachable = 1
 """
-    DOC_ERR_MSG = "Statement is unreachable  [unreachable]"
+    DOC_ERR_MSG = "Statement is unreachable"
 
     # Initialize two workspace folders.
     folder1 = tmpdir.mkdir("folder1")
-    ws1 = Workspace(uris.from_fs_path(str(folder1)), Mock())
-    ws1._config = Config(ws1.root_uri, {}, 0, {})
     folder2 = tmpdir.mkdir("folder2")
-    ws2 = Workspace(uris.from_fs_path(str(folder2)), Mock())
-    ws2._config = Config(ws2.root_uri, {}, 0, {})
 
     # Create configuration file for workspace folder 1.
     mypy_config = folder1.join("mypy.ini")
     mypy_config.write("[mypy]\nwarn_unreachable = True\ncheck_untyped_defs = True")
 
+    ws1 = Workspace(uris.from_fs_path(str(folder1)), Mock())
+    ws1._config = Config(ws1.root_uri, {}, 0, {})
+    ws2 = Workspace(uris.from_fs_path(str(folder2)), Mock())
+    ws2._config = Config(ws2.root_uri, {}, 0, {})
+
     # Initialize settings for both folders.
     plugin.pylsp_settings(ws1._config)
     plugin.pylsp_settings(ws2._config)
@@ -132,6 +125,7 @@ def foo():
     assert len(diags) == 1
     diag = diags[0]
     assert diag["message"] == DOC_ERR_MSG
+    assert diag["code"] == "unreachable"
 
     # Test document in workspace 2 (without mypy.ini configuration)
     doc2 = Document(DOC_URI, ws2, DOC_SOURCE)
@@ -175,10 +169,13 @@ def test_option_overrides(tmpdir, last_diagnostics_monkeypatch, workspace):
         lambda _, p: {"overrides": overrides} if p == "pylsp_mypy" else {},
     )
 
+    config = FakeConfig(uris.to_fs_path(workspace.root_uri))
+    plugin.pylsp_settings(config)
+
     assert not sentinel.exists()
 
     diags = plugin.pylsp_lint(
-        config=FakeConfig(),
+        config=config,
         workspace=workspace,
         document=Document(DOC_URI, workspace, DOC_TYPE_ERR),
         is_saved=False,
@@ -201,13 +198,16 @@ def test_option_overrides_dmypy(last_diagnostics_monkeypatch, workspace):
         else {},
     )
 
-    m = Mock(wraps=lambda a, **_: Mock(returncode=0, **{"stdout.decode": lambda: ""}))
+    m = Mock(wraps=lambda a, **_: Mock(returncode=0, **{"stdout": ""}))
     last_diagnostics_monkeypatch.setattr(plugin.subprocess, "run", m)
 
     document = Document(DOC_URI, workspace, DOC_TYPE_ERR)
 
+    config = FakeConfig(uris.to_fs_path(workspace.root_uri))
+    plugin.pylsp_settings(config)
+
     plugin.pylsp_lint(
-        config=FakeConfig(),
+        config=config,
         workspace=workspace,
         document=document,
         is_saved=False,
@@ -220,10 +220,11 @@ def test_option_overrides_dmypy(last_diagnostics_monkeypatch, workspace):
         "--",
         "--python-executable",
         "/tmp/fake",
-        "--show-column-numbers",
+        "--show-error-end",
+        "--no-error-summary",
         document.path,
     ]
-    m.assert_called_with(expected, stderr=-1, stdout=-1, **windows_flag)
+    m.assert_called_with(expected, capture_output=True, **windows_flag, encoding="utf-8")
 
 
 def test_dmypy_status_file(tmpdir, last_diagnostics_monkeypatch, workspace):
@@ -243,10 +244,13 @@ def test_dmypy_status_file(tmpdir, last_diagnostics_monkeypatch, workspace):
 
     document = Document(DOC_URI, workspace, DOC_TYPE_ERR)
 
+    config = FakeConfig(uris.to_fs_path(workspace.root_uri))
+    plugin.pylsp_settings(config)
+
     assert not statusFile.exists()
 
     plugin.pylsp_lint(
-        config=FakeConfig(),
+        config=config,
         workspace=workspace,
         document=document,
         is_saved=False,
@@ -261,14 +265,10 @@ def foo():
     return
     unreachable = 1
 """
-    DOC_ERR_MSG = "Statement is unreachable  [unreachable]"
+    DOC_ERR_MSG = "Statement is unreachable"
 
     config_sub_paths = [".config"]
 
-    # Initialize workspace.
-    ws = Workspace(uris.from_fs_path(str(tmpdir)), Mock())
-    ws._config = Config(ws.root_uri, {}, 0, {})
-
     # Create configuration file for workspace.
     plugin_config = tmpdir.join("pyproject.toml")
     plugin_config.write(f"[tool.pylsp-mypy]\nenabled = true\nconfig_sub_paths = {config_sub_paths}")
@@ -276,8 +276,14 @@ def foo():
     mypy_config = config_dir.join("mypy.ini")
     mypy_config.write("[mypy]\nwarn_unreachable = True\ncheck_untyped_defs = True")
 
+    # Initialize workspace.
+
+    ws = Workspace(uris.from_fs_path(str(tmpdir)), Mock())
+    ws._config = Config(ws.root_uri, {}, 0, {})
+
     # Update settings for workspace.
-    plugin.pylsp_settings(ws._config)
+    settings = plugin.pylsp_settings(ws._config)
+    ws._config._plugin_settings = _utils.merge_dicts(ws._config._plugin_settings, settings)
 
     # Test document to make sure it uses .config/mypy.ini configuration.
     doc = Document(DOC_URI, ws, DOC_SOURCE)
@@ -285,3 +291,36 @@ def foo():
     assert len(diags) == 1
     diag = diags[0]
     assert diag["message"] == DOC_ERR_MSG
+    assert diag["code"] == "unreachable"
+
+
+def test_config_sub_paths_config_changed(tmpdir, last_diagnostics_monkeypatch):
+    DOC_SOURCE = """
+def foo():
+    return
+    unreachable = 1
+"""
+    DOC_ERR_MSG = "Statement is unreachable"
+
+    # Create configuration file for workspace.
+    config_dir = tmpdir.mkdir(".config")
+    mypy_config = config_dir.join("mypy.ini")
+    mypy_config.write("[mypy]\nwarn_unreachable = True\ncheck_untyped_defs = True")
+
+    config_sub_paths = [".config"]
+
+    # Initialize workspace.
+    ws = Workspace(uris.from_fs_path(str(tmpdir)), Mock())
+    ws._config = Config(ws.root_uri, {}, 0, {})
+
+    # Update settings for workspace.
+    plugin.pylsp_settings(ws._config)
+    ws.update_config({"pylsp": {"plugins": {"pylsp_mypy": {"config_sub_paths": config_sub_paths}}}})
+
+    # Test document to make sure it uses .config/mypy.ini configuration.
+    doc = Document(DOC_URI, ws, DOC_SOURCE)
+    diags = plugin.pylsp_lint(ws._config, ws, doc, is_saved=False)
+    assert len(diags) == 1
+    diag = diags[0]
+    assert diag["message"] == DOC_ERR_MSG
+    assert diag["code"] == "unreachable"
